/*
 * ProfilingWithAspectJ - Profiling with AspectJ
 * Copyright (C) 2007 Christian Schenk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package org.christianschenk.aspectj.profiling;

import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;
import org.aspectj.lang.JoinPoint;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.After;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.Pointcut;

/**
 * This is a simple aspect that can be used for profiling. Subclass a new aspect
 * from here, implement the profile method and define the methods that should be
 * profiled in a pointcut.
 * 
 * @author Christian Schenk
 */
@Aspect
public abstract class AbstractProfilingAspect {

	private static final Logger log = Logger.getLogger("profiling");
	/** Stores the method names and the corresponding execution time */
	private final Map<String, Long> methodTimes;
	/** The top level method which calls the methods that should be profiled */
	private final String topLevelMethod;

	public AbstractProfilingAspect(final String topLevelMethod) {
		this.methodTimes = new HashMap<String, Long>();
		this.topLevelMethod = topLevelMethod;
	}

	@Pointcut
	protected abstract void profile();

	/**
	 * This is the stop watch that takes the time of those methods that should
	 * be profiled.
	 */
	@Around("profile()")
	public void aroundProfileMethods(final ProceedingJoinPoint thisJoinPoint) throws Throwable {
		final long start, end;
		start = System.nanoTime();
		thisJoinPoint.proceed();
		end = System.nanoTime();
		this.methodTimes.put(this.getJoinPointName(thisJoinPoint), ((end - start) / 1000) / 1000);
	}

	/**
	 * Prints the gathered statistics once the top level method is finished.
	 */
	@After("profile()")
	public void afterTopLevelMethod(final JoinPoint joinPoint) {
		if (this.topLevelMethod.equals(this.getJoinPointName(joinPoint)) == false) return;

		final long overallTime = this.methodTimes.get(this.topLevelMethod);
		log.debug("Results for " + this.topLevelMethod + " (" + overallTime + "ms)");
		for (final String method : this.methodTimes.keySet()) {
			if (method.equals(this.topLevelMethod)) continue;
			final long methodTime = this.methodTimes.get(method);
			final double percent = Math.round(((double) methodTime / (double) overallTime) * 100);
			log.debug("Method: " + method + "() took " + methodTime + "ms (" + percent + "%)");
		}
	}

	private final String getJoinPointName(final JoinPoint joinPoint) {
		return joinPoint.getThis().getClass().getSimpleName() + "." + joinPoint.getSignature().getName();
	}
}